#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>

#include "lwip/opt.h"
#include "lwip/init.h"

#include "lwip/mem.h"
#include "lwip/memp.h"
#include "lwip/sys.h"

#include "lwip/stats.h"

#include "lwip/inet_chksum.h"
#include "lwip/tcpip.h"
#include "lwip/sockets.h"

#include "netif/tapif.h"
#include "netif/loopif.h"
#include "netif/tcpdump.h"

#include "lwip/ip_addr.h"

#include "arch/perf.h"
#include "arch/sys_arch.h"

#define NUM_THREADS 10
#define SOCKETS_PER_THREAD 40

#define PERF_LOOPIF 0
#define PERF_TAPIF (PERF_LOOPIF==0)

#if PERF_LOOPIF
#ifndef LWIP_HAVE_LOOPIF
#error "Must use loopback interface."
#endif
#endif /*PERF_LOOPIF*/

char name[] = "PERFTEST5";
char filename[] = "dummy.csv";
#define INP_NAME_SIZE 100
char inputname[INP_NAME_SIZE];
char perror_string[100];

/* (manual) host IP configuration */
static struct ip_addr ipaddr, netmask, gw;

/* nonstatic debug cmd option, exported in lwipopts.h */
unsigned char debug_flags;

#if PERF_LOOPIF
struct netif loopif;
#endif

#if PERF_TAPIF
struct netif tapif;
#endif

struct priority_arg {
  int socket;
  u16_t send_port;
  int send_tos;
  u16_t recv_port;
  sys_sem_t sem;
  int packets;
  u16_t own_port;
};  

struct thread_arg {
  int sockets[SOCKETS_PER_THREAD];
  u16_t send_ports[SOCKETS_PER_THREAD];
  u16_t recv_ports[SOCKETS_PER_THREAD];
  sys_sem_t sem;
  int packets;
};  

struct big {
  int numbers[350];
  int send_value;
};

static void
tcpip_init_done(void *arg)
{
  sys_sem_t *sem;
  sem = arg;
  sys_sem_signal(*sem);
}

/*
 *
 * SOCKET THREAD
 *
 */

static void
socket_thread(void *arg)
{
  int i,j;
  struct sockaddr_in remote_addr;
  struct sockaddr_in my_addr;

  /* for socket send */
  struct big big_send;
  int send_count = 0;
  unsigned int send_flags = 0;

  /* for getsockopt */
  int optval = 0;
  socklen_t optlen; 

  struct thread_arg * ta;
  ta = arg;

  /* REMOTE ADDRESS - PORT IS THREAD-SPECIFIC */
  remote_addr.sin_family = AF_INET;
  #if PERF_TAPIF
  /*remote_addr.sin_addr.s_addr = inet_addr("127.0.0.1");*/
  remote_addr.sin_addr.s_addr = inet_addr("129.241.187.94");
  #endif
  #if PERF_LOOPIF
  remote_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  #endif  
  memset(remote_addr.sin_zero,'\0',sizeof remote_addr.sin_zero);

  /* MY_ADDR FOR BIND() - PORT IS THREAD-SPECIFIC */
  my_addr.sin_family = AF_INET;
  #if PERF_TAPIF
  my_addr.sin_addr.s_addr = inet_addr("192.168.0.2");
  /*my_addr.sin_addr.s_addr = inet_addr("127.0.0.1");*/
  #endif
  #if PERF_LOOPIF
  my_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  #endif
  memset(my_addr.sin_zero,'\0',sizeof my_addr.sin_zero);

  /* SOCKET(), BIND() */
  for(i=0; i<SOCKETS_PER_THREAD; i++){
    ta->sockets[i] = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(ta->sockets[i] == -1){
      sprintf(perror_string,"sockets[%d]: socket",ta->sockets[i]);
      perror(perror_string);
      return;
    }

    my_addr.sin_port = htons(ta->send_ports[i]);

    if( lwip_bind(ta->sockets[i], (struct sockaddr *) &my_addr, (socklen_t) sizeof(my_addr) ) == -1 ){
      sprintf(perror_string,"sockets[%d]: bind",ta->sockets[i]);  
      perror(perror_string);      
      return;
    }
  }

  /* Signal main thread that init is complete */
  sys_sem_signal(ta->sem);
  sys_sem_wait(ta->sem);

  send_flags = 0;
  for(i=0; i<SOCKETS_PER_THREAD; i++){
    remote_addr.sin_port = htons(ta->recv_ports[i]);
    for(j = 0; j < ta->packets ; j++ ){
      big_send.send_value = j;
      send_count = lwip_sendto(ta->sockets[i], &big_send, sizeof(big_send), send_flags, 
                                (struct sockaddr*)&remote_addr, (socklen_t) sizeof(remote_addr) );
      if( send_count == -EWOULDBLOCK){
        printf("%s: Socket%d EWOULDBLOCK, breaking send loop\n", name, ta->sockets[i]);        
        break;
      }
      else if(send_count == -1){
        sprintf(perror_string,"sockets[%d]: sendto %d",ta->sockets[i],j);
        perror(perror_string);
        return;
      } 
      else {
        /*OK*/
      }
    }
  }

  /* ASSERT NO ERRORS OCCURED */
  optlen = sizeof optval;
  for(i=0; i<SOCKETS_PER_THREAD; i++){
    if( lwip_getsockopt(ta->sockets[i], SOL_SOCKET, SO_ERROR, &optval, &optlen ) == -1){
      sprintf(perror_string,"sockets[%d]: getsockopt",ta->sockets[i]);
      perror(perror_string);
      return;
    }
    LWIP_ASSERT("PBMS: No errors in socket after test run", (optval == ERR_OK) );
  }
  printf("%s: Sockets %d-%d finished.\n", name, ta->sockets[0], ta->sockets[SOCKETS_PER_THREAD-1]);
} 

/*
 *
 * PRIORITY THREAD
 * TODO: do some receives in this as well
 */

static void
priority_thread(void *arg)
{
  int i;
  struct sockaddr_in remote_addr;
  struct sockaddr_in my_addr;

  /* for socket send */
  struct big big_send;
  int send_count = 0;
  unsigned int send_flags = 0;

  /* for socket receive */
  int recv_count = 0;
  int recv_value = 0;
  unsigned int recv_flags = 0;
  socklen_t my_addr_size = 0;

  int ioctl_arg = 0;

  /* for getsockopt */
  int optval = 0;
  socklen_t optlen;   

  struct priority_arg * ta;
  ta = arg;
  
  /* SOCKET() */
  ta->socket = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(ta->socket == -1){
    perror("Priority thread: socket");
    return;
  }
    
  /* BIND() */
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(ta->send_port);
  #if PERF_TAPIF
  my_addr.sin_addr.s_addr = inet_addr("192.168.0.2");
  /*my_addr.sin_addr.s_addr = inet_addr("127.0.0.1");*/
  #endif
  #if PERF_LOOPIF
  my_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  #endif
  memset(my_addr.sin_zero,'\0',sizeof my_addr.sin_zero);
  if( lwip_bind(ta->socket, (struct sockaddr *) &my_addr, (socklen_t) sizeof(my_addr) ) == -1 ){
    perror("Priority thread: bind");
    return;
  }

  /* SETSOCKOPT() - TOS */
  if( lwip_setsockopt(ta->socket, IPPROTO_IP, IP_TOS, &(ta->send_tos), (socklen_t) sizeof(ta->send_tos) ) == -1){
    perror("Priority thread: setsockopt");
    return;
  }  

  /* REMOTE ADDRESS */
  remote_addr.sin_family = AF_INET;
  remote_addr.sin_port = htons(ta->recv_port);
  #if PERF_TAPIF
  /*remote_addr.sin_addr.s_addr = inet_addr("127.0.0.1");*/
  remote_addr.sin_addr.s_addr = inet_addr("129.241.187.94");
  #endif
  #if PERF_LOOPIF
  remote_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  #endif  
  memset(remote_addr.sin_zero,'\0',sizeof remote_addr.sin_zero);

  /* Signal main thread that init is complete */
  sys_sem_signal(ta->sem);
  sys_sem_wait(ta->sem);

  /* Non-blocking receive */
  send_flags = 0;
  recv_flags = 0;      
  ioctl_arg = 1;       
  lwip_ioctl(ta->socket,FIONBIO,&ioctl_arg);      

  for(i = 0; i < ta->packets ; i++ ){
    big_send.send_value = i;
    send_count = lwip_sendto(ta->socket, &big_send, sizeof(big_send), send_flags, 
                              (struct sockaddr*)&remote_addr, (socklen_t) sizeof(remote_addr) );

    if( send_count == -EWOULDBLOCK){
      printf("%s: Socket%d EWOULDBLOCK, breaking send loop\n", name, ta->socket);        
      break;
    }
    else if(send_count == -1){
      sprintf(perror_string,"Priority thread: sendto %d",i);
      perror(perror_string);
      return;
    } 
    else {
      /* OK */
    }
  }

  /*sleep(1);*/

  /* TODO: RX in request-response pattern */ 
  my_addr_size = sizeof(my_addr);
  for(i = 0; i<ta->packets ; i++){
    recv_count = lwip_recvfrom(ta->socket, (void *) &recv_value, sizeof(recv_value), recv_flags, 
                                (struct sockaddr*) &my_addr, &my_addr_size );
    if( recv_count == -EWOULDBLOCK ){
      printf("%s: Socket%d EWOULDBLOCK, breaking recv loop\n", name, ta->socket);        
      break;
    }
    else if( recv_count == -1 ){
      sprintf(perror_string,"Priority thread: recvfrom %d",i);
      perror(perror_string);
      return;
    }
    else {
      /* OK */
    }
  }

  /* ASSERT NO ERRORS OCCURED */
  optlen = sizeof(optval);
  if( lwip_getsockopt(ta->socket, SOL_SOCKET, SO_ERROR, &optval, &optlen ) == -1){
    perror("Priority thread: getsockopt");
    return;
  }
  LWIP_ASSERT("PBMS: No errors in socket after test run", (optval == ERR_OK) );

  printf("%s: Priority socket%d finished.\n", name, ta->socket);
} 

void test_mbox_capacity(){
  sys_mbox_t mbox;
  int mbox_posts = 0;
  int netbuf_allocs = 0;
  int free_counter = 0;
  err_t post_return;
  struct netbuf *buf = NULL;

  mbox = sys_mbox_new(TCPIP_MBOX_SIZE);
  do {
    buf = memp_malloc(MEMP_NETBUF);
    if (buf == NULL) {
      printf("%s :: netbuf allocation error after %d allocations\n", name, netbuf_allocs);
      break;
    } 
    else {
      netbuf_allocs++;
    }

    post_return = sys_mbox_trypost(mbox, buf);  
    if( post_return == ERR_OK ){
      mbox_posts++;
    }
    else {
      memp_free(MEMP_NETBUF,(void *)buf);
      LWIP_ASSERT("PBMS: message box capacity as expected", (mbox_posts == (SYS_MBOX_SIZE-1)) );
      LWIP_ASSERT("PBMS: trypost returned ERR_MEM", (post_return == ERR_MEM) );
      printf("%s: mbox test run: mbox full after %d posts \n",name ,mbox_posts );      
      break;
    } 
  } while(1);

  printf("%s: mbox test run: freed NETBUF ",name );
  for(free_counter = 0; free_counter < mbox_posts; free_counter++){
    sys_mbox_fetch(mbox,(void *)&buf);
    memp_free(MEMP_NETBUF,(void *)buf);
    printf("#%d, ",free_counter);
  }
  printf("mbox test run done. \n");
}

void init(){
  sys_sem_t sem;
  struct in_addr inaddr;
  char ip_str[16] = {0}, nm_str[16] = {0}, gw_str[16] = {0};
  
  /* use debug flags defined by debug.h */
  debug_flags |= (LWIP_DBG_ON|LWIP_DBG_TRACE|LWIP_DBG_STATE|LWIP_DBG_FRESH|LWIP_DBG_HALT);

  lwip_init();
  stats_init();
 
  netif_init();
  sem = sys_sem_new(0);
  tcpip_init(tcpip_init_done, &sem);
  sys_sem_wait(sem);
  sys_sem_free(sem);

#if PERF_TAPIF
  IP4_ADDR(&gw, 192,168,0,1);
  IP4_ADDR(&netmask, 255,255,255,0);
  IP4_ADDR(&ipaddr, 192,168,0,2);

  printf("%s :: Using tap interface.\n", name);
  netif_set_default(netif_add(&tapif,&ipaddr, &netmask, &gw, NULL, tapif_init,
                  tcpip_input));
  netif_set_up(&tapif);
#endif

#if PERF_LOOPIF
  IP4_ADDR(&gw, 127,0,0,1);
  IP4_ADDR(&ipaddr, 127,0,0,1);
  IP4_ADDR(&netmask, 255,0,0,0);

  printf("%s :: Using loopback interface.\n", name);
  netif_set_default(netif_add(&loopif, &ipaddr, &netmask, &gw, NULL, loopif_init,
        tcpip_input));
  netif_set_up(&loopif);        
#endif

  inaddr.s_addr = ipaddr.addr;
  strncpy(ip_str,inet_ntoa(inaddr),sizeof(ip_str));
  inaddr.s_addr = netmask.addr;
  strncpy(nm_str,inet_ntoa(inaddr),sizeof(nm_str));
  inaddr.s_addr = gw.addr;
  strncpy(gw_str,inet_ntoa(inaddr),sizeof(gw_str));
  printf("Host at %s mask %s gateway %s\n", ip_str, nm_str, gw_str);

  printf("%s :: System initialized.\n", name);
  
  test_mbox_capacity();
}

int
main(int argc, char **argv)
{
  int i,j;
  const int sleeptime = 20; 
  const int packets = 40000;
  const int pri_packets = 10000;
  int packets_per_socket = packets /(NUM_THREADS * SOCKETS_PER_THREAD);

  struct thread_arg threads[NUM_THREADS];
  struct priority_arg pri_arg;

  system("clear");
  
  memset(inputname,'\0',INP_NAME_SIZE);
  if (argc == 1){
    printf("Optional command line argument: Debug print file name.\n");    
    printf("Using default name: %s\n",filename);
  }
  else if (argc == 2) {
    strncpy(inputname, argv[1], INP_NAME_SIZE-1);
    printf("Debug perf print file name: %s\n",inputname);
  }
  else {
    printf("Command line argument: Optional single file name only. Exiting.\n");
    exit(0);
  }

  pri_arg.socket = -1;
  pri_arg.send_port = 10000;
  pri_arg.send_tos = IPTOS_PREC_PRIORITY;
  pri_arg.recv_port = 20000;
  pri_arg.sem = sys_sem_new(0);
  pri_arg.packets = pri_packets;

  for(i=0; i<NUM_THREADS ; i++ ){
    for(j=0; j<SOCKETS_PER_THREAD ; j++ ){
      threads[i].sockets[j] = -1;
      threads[i].send_ports[j] = 10001 + i;
      threads[i].recv_ports[j] = 20001 + i;  
    }
    threads[i].sem = sys_sem_new(0);
    threads[i].packets = packets_per_socket;
  }

  init(); 

  #ifdef PERF
  /*  perf_init("dummy.csv");*/
  if(argc==1){
    perf_init(filename);
  }
  else if (argc==2) {
    perf_init(inputname);  
  }
  else {
    printf("perf init error. Exit.\n");
    exit(0);
  }
  #endif /* PERF */

  sleep(sleeptime);

  /* Thread initialization */
  sys_thread_new("priority_thread", priority_thread, (void *)&pri_arg, DEFAULT_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
  printf("%s: Main: Waiting for priority thread init.\n", name);
  sys_sem_wait(pri_arg.sem);

  for(i=0; i<NUM_THREADS ; i++ ){
    sys_thread_new("socket_thread", socket_thread, (void *)&threads[i], DEFAULT_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
    printf("%s: Main: Waiting for thread %d init.\n", name, i);
    sys_sem_wait(threads[i].sem);
  }

  /* TX START SYNCHRONIZATION */
  printf("%s: Main: Routine threads begin TX: ", name);
  for(i=0; i<NUM_THREADS ; i++ ){
    printf("%d ",i);
    sys_sem_signal(threads[i].sem);
  }
  printf(" :: All started\n");
  printf("%s: Main: Signalling priority thread to begin transmission.\n", name);
  sys_sem_signal(pri_arg.sem);  

  sleep(20);

  /* TODO: TX END SYNCHRONIZATION ? */

  stats_display(); 
  printf("\a");
  return 0;
}


