/*
 * Copyright (c) 2001-2003 Swedish Institute of Computer Science.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>

#include "lwip/opt.h"
#include "lwip/init.h"

#include "lwip/mem.h"
#include "lwip/memp.h"
#include "lwip/sys.h"

#include "lwip/stats.h"

#include "lwip/inet_chksum.h"
#include "lwip/tcpip.h"
#include "lwip/sockets.h"

#include "netif/tapif.h"
#include "netif/loopif.h"
#include "netif/tcpdump.h"

#include "lwip/ip_addr.h"
#include "lwip/debug.h"
#include "arch/perf.h"
#include "arch/sys_arch.h"

/* (manual) host IP configuration */
static struct ip_addr ipaddr, netmask, gw;

/* ping out destination cmd option */
static unsigned char ping_flag;
static struct ip_addr ping_addr;

/* nonstatic debug cmd option, exported in lwipopts.h */
unsigned char debug_flags;

/** @todo add options for selecting netif, starting DHCP client etc */
static struct option longopts[] = {
  /* turn on debugging output (if build with LWIP_DEBUG) */
  {"debug", no_argument,        NULL, 'd'},
  /* help */
  {"help", no_argument, NULL, 'h'},
  /* gateway address */
  {"gateway", required_argument, NULL, 'g'},
  /* ip address */
  {"ipaddr", required_argument, NULL, 'i'},
  /* netmask */
  {"netmask", required_argument, NULL, 'm'},
  /* ping destination */
  {"ping",   required_argument, NULL, 'p'},
  /* new command line options go here! */
  {NULL,   0,                 NULL,  0}
};
#define NUM_OPTS ((sizeof(longopts) / sizeof(struct option)) - 1)

void usage(void)
{
  unsigned char i;
   
  printf("options:\n");
  for (i = 0; i < NUM_OPTS; i++) {
    printf("-%c --%s\n",longopts[i].val, longopts[i].name);
  }
}

static void
tcp_debug_timeout(void *data)
{
#if TCP_DEBUG
  tcp_debug_print_pcbs();
#endif /* TCP_DEBUG */
  sys_timeout(5000, tcp_debug_timeout, NULL);
}

static void
tcpip_init_done(void *arg)
{
  sys_sem_t *sem;
  sem = arg;
  sys_sem_signal(*sem);
}


struct netif netif;
#if LWIP_HAVE_LOOPIF
struct netif loopif;
#endif

static void
main_thread(void *arg)
{
  sys_sem_t sem;

  int socket1 = -1;
  const u16_t s1_port = 10000;
  struct sockaddr_in s1_addr;
  int s1_tos;   
  int tos_optval;   
  socklen_t tos_optlen;

  int socket2 = -1;
  const u16_t s2_port = 20000;
  struct sockaddr_in s2_addr;
  
  int send_count;
  unsigned int send_flags;
  struct sockaddr_in dest_addr;
  socklen_t dest_len;
  int send_i = 0;

  int recv_count;
  int recv_value;
  unsigned int recv_flags;
  struct sockaddr send_addr;
  socklen_t send_addr_size;

  /* lwIP INIT */
  netif_init();
  sem = sys_sem_new(0);
  tcpip_init(tcpip_init_done, &sem);
  sys_sem_wait(sem);
  sys_sem_free(sem);
  printf("TCP/IP initialized.\n");
/*netif_set_default(netif_add(&netif,&ipaddr, &netmask, &gw, NULL, tapif_init,
                  tcpip_input));
  netif_set_up(&netif);  */
 
#if LWIP_HAVE_LOOPIF  
  IP4_ADDR(&gw, 127,0,0,1);
  IP4_ADDR(&ipaddr, 127,0,0,1);
  IP4_ADDR(&netmask, 255,0,0,0);
  netif_set_default(netif_add(&loopif, &ipaddr, &netmask, &gw, NULL, loopif_init,
        tcpip_input));
  netif_set_up(&loopif);        
#endif

  /* SOCKET() */
  socket1 = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(socket1 == -1){
    perror("socket1");
  }

  /* BIND() */
  s1_addr.sin_family = AF_INET;
  s1_addr.sin_port = htons(s1_port);
  s1_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(s1_addr.sin_zero,'\0',sizeof s1_addr.sin_zero);
  if( lwip_bind(socket1, (struct sockaddr *) &s1_addr, (socklen_t) sizeof(s1_addr) ) == -1 ){
    perror("bind");  
  }

  /* SETSOCKOPT() */
  s1_tos = IPTOS_PREC_PRIORITY;
  if( lwip_setsockopt(socket1, IPPROTO_IP, IP_TOS, &s1_tos, (socklen_t) sizeof(s1_tos) ) == -1){
    perror("setsockopt");
  }  

  /* GETSOCKOPT() */
  tos_optlen = sizeof(tos_optval);
  if( lwip_getsockopt(socket1, IPPROTO_IP, IP_TOS, &tos_optval, &tos_optlen ) == -1){
    perror("getsockopt");
  }
  LWIP_DEBUGF(SERVER_DEBUG, ("Server: Socket%d TOS: %d \n",socket1 , tos_optval) );

  /* SOCKET 2 */
  socket2 = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);  
  if(socket2 == -1){
    perror("socket2");
  }
  s2_addr.sin_family = AF_INET;
  s2_addr.sin_port = htons(s2_port);
  s2_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(s2_addr.sin_zero,'\0',sizeof s2_addr.sin_zero);
  if( lwip_bind(socket2, (struct sockaddr *) &s2_addr, (socklen_t) sizeof(s2_addr) ) == -1 ){
    perror("bind");  
  }
    
  /*
  LWIP_DEBUGF(SERVER_DEBUG, ("TOS Server :: sleep 5\n") );
  sys_msleep(5000);
  */

  /* SEND FROM SOCKET1 TO SOCKET2 */
  send_flags = 0;
  dest_addr.sin_len = sizeof(dest_addr);
  memset(dest_addr.sin_zero,'\0',sizeof dest_addr.sin_zero);
  dest_addr.sin_port = htons(s2_port);
  dest_addr.sin_family = AF_INET;
  dest_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  /*dest_addr.sin_addr.s_addr = inet_addr("127.0.0.1");*/
  dest_len = sizeof(dest_addr);

  for(send_i = 0; send_i<10 ; send_i++){
    send_count = lwip_sendto(socket1, &send_i, sizeof(send_i), send_flags, (struct sockaddr*)&dest_addr, dest_len );
    if( send_count == -1){
      perror("sendto");
    }
    LWIP_DEBUGF(SERVER_DEBUG, ("Server: send #%d: Socket%d sent value: %d (size: %d bytes).\n", send_i, socket1 , send_i, send_count) );
  }

  /* TODO: test MSG_PEEK */
  recv_flags = 0;
  send_addr_size = sizeof(send_addr);
  for(send_i = 0; send_i<10 ; send_i++){
    recv_count = lwip_recvfrom(socket2, (void *) &recv_value, sizeof(recv_value), recv_flags, &send_addr, &send_addr_size );
    if( recv_count == -1 ){
      perror("recvfrom");
    }
    LWIP_DEBUGF(SERVER_DEBUG, ("Server: recv #%d: Socket%d received value: %d (size: %d bytes).\n",send_i, socket2 , recv_value, recv_count) );    
  }




  
  /* Block for ever. */
  sem = sys_sem_new(0);
  sys_sem_wait(sem);
}

/*-----------------------------------------------------------------------------------*/

int
main(int argc, char **argv)
{
  struct in_addr inaddr;
  int ch;
  char ip_str[16] = {0}, nm_str[16] = {0}, gw_str[16] = {0};

  /* startup defaults (may be overridden by one or more opts) */
  IP4_ADDR(&gw, 192,168,0,1);
  IP4_ADDR(&netmask, 255,255,255,0);
  IP4_ADDR(&ipaddr, 192,168,0,2);
  
  ping_flag = 0;
  /* use debug flags defined by debug.h */
  debug_flags = LWIP_DBG_ON;
  
  while ((ch = getopt_long(argc, argv, "dhg:i:m:p:", longopts, NULL)) != -1) {
    switch (ch) {
      case 'd':
        debug_flags |= (LWIP_DBG_ON|LWIP_DBG_TRACE|LWIP_DBG_STATE|LWIP_DBG_FRESH|LWIP_DBG_HALT);
        break;
      case 'h':
        usage();
        exit(0);
        break;
      case 'g':
        inet_aton(optarg, &inaddr);
        gw.addr = inaddr.s_addr;
        break;
      case 'i':
        inet_aton(optarg, &inaddr);
        ipaddr.addr = inaddr.s_addr;
        break;
      case 'm':
        inet_aton(optarg, &inaddr);
        netmask.addr = inaddr.s_addr;
        break;
      case 'p':
        ping_flag = !0;
        inet_aton(optarg, &inaddr);
        /* lwip inet.h oddity workaround */
        ping_addr.addr = inaddr.s_addr; 
        strncpy(ip_str,inet_ntoa(inaddr),sizeof(ip_str));
        printf("Using %s to ping\n", ip_str);
        break;
      default:
        usage();
        break;
    }
  }
  argc -= optind;
  argv += optind;

  inaddr.s_addr = ipaddr.addr;
  strncpy(ip_str,inet_ntoa(inaddr),sizeof(ip_str));
  inaddr.s_addr = netmask.addr;
  strncpy(nm_str,inet_ntoa(inaddr),sizeof(nm_str));
  inaddr.s_addr = gw.addr;
  strncpy(gw_str,inet_ntoa(inaddr),sizeof(gw_str));
  printf("Host at %s mask %s gateway %s\n", ip_str, nm_str, gw_str);

#ifdef PERF
  perf_init("/tmp/simhost.perf");
#endif /* PERF */

  lwip_init();
  
  printf("System initialized.\n");
    
  sys_thread_new("main_thread", main_thread, NULL, DEFAULT_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
  pause();
  return 0;
}

#if 0
  printf("dest_addr.sin_len = %d || ",dest_addr.sin_len);
  printf("dest_addr.sin_family = %d || ",dest_addr.sin_family);
  printf("dest_addr.sin_addr.socket1_addr = %"U32_F" \n ",dest_addr.sin_addr.socket1_addr);
  printf("&dest_addr (to) = %p || ", (void *) &dest_addr);
  printf("dest_len (to_len) = %"U32_F" || ",dest_len);  
  printf("dest_addr.sin_zero = ");
  for(array_i = 0; array_i<8 ; array_i++){ printf(" %d",dest_addr.sin_zero[array_i]); }
  printf("\n");
#endif



