#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>

#include "lwip/opt.h"
#include "lwip/init.h"

#include "lwip/mem.h"
#include "lwip/memp.h"
#include "lwip/sys.h"

#include "lwip/stats.h"

#include "lwip/inet_chksum.h"
#include "lwip/tcpip.h"
#include "lwip/sockets.h"

#include "netif/tapif.h"
#include "netif/loopif.h"
#include "netif/tcpdump.h"

#include "lwip/ip_addr.h"
#include "lwip/debug.h"
#include "arch/perf.h"
#include "arch/sys_arch.h"

#ifndef LWIP_HAVE_LOOPIF
#error "Must use loopback interface."
#endif

#define THREAD_PAIRS 2

char name[] = "TEST 4";

/* (manual) host IP configuration */
static struct ip_addr ipaddr, netmask, gw;

/* nonstatic debug cmd option, exported in lwipopts.h */
unsigned char debug_flags;

struct netif loopif;

struct thread_arg {
  int send_socket;
  u16_t send_port;
  int send_tos;
  int recv_socket;
  u16_t recv_port;
  int recv_tos;
  sys_sem_t sem;
  int packets;
  int iterations;
};  
  
static void
tcpip_init_done(void *arg)
{
  sys_sem_t *sem;
  sem = arg;
  sys_sem_signal(*sem);
}

static void
socket_thread(void *arg)
{
  int i,j,k;

  struct sockaddr_in send_addr;
  struct sockaddr_in recv_addr;

  /* for socket send */
  int send_count = 0;
  int send_value = 0;
  unsigned int send_flags = 0;

  /* for socket receive */
  int recv_count = 0;
  int recv_value = 0;
  unsigned int recv_flags = 0;
  socklen_t send_addr_size = 0;

  int ioctl_arg = 0;

  /* for getsockopt */
  int optval = 0;
  socklen_t optlen;   

  struct thread_arg * ta;
  ta = arg;
  
  optlen = sizeof(optval);

  /* SOCKET() */
  ta->send_socket = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(ta->send_socket == -1){
    perror("socket");
  }
  ta->recv_socket = lwip_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);  
  if(ta->recv_socket == -1){
    perror("socket");
  }
    
  /* BIND() */
  LWIP_ASSERT("PBMS: Using proper loopback address", (htonl(INADDR_LOOPBACK) == inet_addr("127.0.0.1")) );
  send_addr.sin_family = AF_INET;
  send_addr.sin_port = htons(ta->send_port);
  send_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(send_addr.sin_zero,'\0',sizeof send_addr.sin_zero);
  if( lwip_bind(ta->send_socket, (struct sockaddr *) &send_addr, (socklen_t) sizeof(send_addr) ) == -1 ){
    perror("bind");  
  }
  recv_addr.sin_family = AF_INET;
  recv_addr.sin_port = htons(ta->recv_port);
  recv_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(recv_addr.sin_zero,'\0',sizeof recv_addr.sin_zero);
  if( lwip_bind(ta->recv_socket, (struct sockaddr *) &recv_addr, (socklen_t) sizeof(recv_addr) ) == -1 ){
    perror("bind");  
  }

  /* SETSOCKOPT() - TOS */
  if( lwip_setsockopt(ta->send_socket, IPPROTO_IP, IP_TOS, &(ta->send_tos), (socklen_t) sizeof(ta->send_tos) ) == -1){
    perror("setsockopt");
  }  
  if( lwip_setsockopt(ta->recv_socket, IPPROTO_IP, IP_TOS, &(ta->recv_tos), (socklen_t) sizeof(ta->recv_tos) ) == -1){
    perror("setsockopt");
  }  

  /* Note: PBMS has only nonblocking send */

/*  ioctl_arg = 1;   
  if( lwip_ioctl(ta->send_socket,FIONBIO,&ioctl_arg) == -1){
    perror("ioctl");
  }    */

  /* Signal main thread that init is complete */
  
  LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket pair (%d,%d) Signalling main thread.\n", name, ta->send_socket, ta->recv_socket) );        
  sys_sem_signal(ta->sem);
  sys_sem_wait(ta->sem);
  LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket pair (%d,%d) starting transmission.\n", name, ta->send_socket, ta->recv_socket) );        

  for(i = 0; i < ta->iterations; i++ ){
    LWIP_DEBUGF(SERVER_DEBUG_SEND, ("%s: Send loop: (Iteration, send number, socket number, sent value, sendto() return value): \n", name) );
    LWIP_DEBUGF(SERVER_DEBUG_SEND, ("\t") );
    send_flags = 0;
    for(j = 0; j < ta->packets ; j++ ){
      send_value = j + 10*i;
      send_count = lwip_sendto(ta->send_socket, &send_value, sizeof(send_value), send_flags, 
                                (struct sockaddr*)&recv_addr, (socklen_t) sizeof(recv_addr) );
/*      LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket pair (%d,%d): send_count: %d.\n", name, ta->send_socket, ta->recv_socket, send_count) );*/

      if( send_count == -EWOULDBLOCK){
        LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket%d EWOULDBLOCK, breaking send loop\n", name, ta->send_socket) );        
        break;
      }
      else if(send_count == -1){
        perror("sendto");
      } 
      else {
        LWIP_DEBUGF(SERVER_DEBUG_SEND, ("(%d,%d,%d,%d,%d) ", i, j, ta->send_socket, send_value, send_count) );
      }
    }
    LWIP_DEBUGF(SERVER_DEBUG_SEND, ("\n\n") );

    LWIP_DEBUGF(SERVER_DEBUG_SEND, ("%s: Thread sleeping to let all sends be processed.\n",name));
    sleep(1);

    /* NONBLOCKING RECEIVE */
    /* test receive using MSG_DONTWAIT every other iteration */
    if( i%2 ){
      ioctl_arg = 0;   
      recv_flags = MSG_DONTWAIT;
    }
    else {
      ioctl_arg = 1;       
      recv_flags = 0;      
    }
    lwip_ioctl(ta->recv_socket,FIONBIO,&ioctl_arg);      

    send_addr_size = sizeof(send_addr);

    LWIP_DEBUGF(SERVER_DEBUG_RECV, ("%s: Receive loop: (Iteration, receive number, socket number, received value, recvfrom() return value): \n", name) );
    LWIP_DEBUGF(SERVER_DEBUG_RECV, ("\t") );
    for(k = 0; k<ta->packets ; k++){
      recv_count = lwip_recvfrom(ta->recv_socket, (void *) &recv_value, sizeof(recv_value), recv_flags, 
                                  (struct sockaddr*) &send_addr, &send_addr_size );
      if( recv_count == -EWOULDBLOCK ){
        LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket%d EWOULDBLOCK, breaking recv loop\n", name, ta->recv_socket) );        
        break;
      }
      else if( recv_count == -1 ){
        perror("recvfrom");
      }
      else {
        LWIP_DEBUGF(SERVER_DEBUG_RECV, ("(%d,%d,%d,%d,%d) ", i, k, ta->recv_socket, recv_value, recv_count) );    
      }
    }

    /* ASSERT NO ERRORS OCCURED */
    if( lwip_getsockopt(ta->send_socket, SOL_SOCKET, SO_ERROR, &optval, &optlen ) == -1){
      perror("getsockopt");
    }
/*    LWIP_DEBUGF(SERVER_DEBUG, ("PBMS: Errors in send_socket after test run: %d \n", optval) );*/
    LWIP_ASSERT("PBMS: No errors in send_socket after test run", (optval == ERR_OK) );
    if( lwip_getsockopt(ta->recv_socket, SOL_SOCKET, SO_ERROR, &optval, &optlen ) == -1){
      perror("getsockopt");
    }
/*    LWIP_DEBUGF(SERVER_DEBUG, ("PBMS: Errors in recv_socket after test run: %d \n", optval) );*/
    LWIP_ASSERT("PBMS: No errors in recv_socket after test run", (optval == ERR_OK) );

  } /* end for(iterations) */

  LWIP_DEBUGF(SERVER_DEBUG, ("%s: Socket pair (%d,%d) finished.\n", name, ta->send_socket, ta->recv_socket) );
} 

void test_mbox_capacity(){
  sys_mbox_t mbox;
  int mbox_posts = 0;
  int netbuf_allocs = 0;
  int free_counter = 0;
  err_t post_return;
  struct netbuf *buf = NULL;

  mbox = sys_mbox_new(TCPIP_MBOX_SIZE);
  do {
    buf = memp_malloc(MEMP_NETBUF);
    if (buf == NULL) {
      printf("%s :: netbuf allocation error after %d allocations\n", name, netbuf_allocs);
      break;
    } 
    else {
      netbuf_allocs++;
    }

    post_return = sys_mbox_trypost(mbox, buf);  
    if( post_return == ERR_OK ){
      mbox_posts++;
    }
    else {
      memp_free(MEMP_NETBUF,(void *)buf);
      LWIP_ASSERT("PBMS: message box capacity as expected", (mbox_posts == (SYS_MBOX_SIZE-1)) );
      LWIP_ASSERT("PBMS: trypost returned ERR_MEM", (post_return == ERR_MEM) );
      printf("%s: mbox test run: mbox full after %d posts \n",name ,mbox_posts );      
      break;
    } 
  } while(1);

  printf("%s: mbox test run: freed NETBUF ",name );
  for(free_counter = 0; free_counter < mbox_posts; free_counter++){
    sys_mbox_fetch(mbox,(void *)&buf);
    memp_free(MEMP_NETBUF,(void *)buf);
    printf("#%d, ",free_counter);
  }
  printf("mbox test run done. \n");
}

void init(){
  sys_sem_t sem;
  
  /* use debug flags defined by debug.h */
  debug_flags |= (LWIP_DBG_ON|LWIP_DBG_TRACE|LWIP_DBG_STATE|LWIP_DBG_FRESH|LWIP_DBG_HALT);
  lwip_init();
  stats_init();
  netif_init();
  sem = sys_sem_new(0);
  tcpip_init(tcpip_init_done, &sem);
  sys_sem_wait(sem);
  sys_sem_free(sem);

  IP4_ADDR(&gw, 127,0,0,1);
  IP4_ADDR(&ipaddr, 127,0,0,1);
  IP4_ADDR(&netmask, 255,0,0,0);
  netif_set_default(netif_add(&loopif, &ipaddr, &netmask, &gw, NULL, loopif_init,
        tcpip_input));
  netif_set_up(&loopif);        

  system("clear");
  printf("%s :: System initialized.\n", name);
  
  test_mbox_capacity();
}

int
main(int argc, char **argv)
{
  int i;
  const int packets = 10;
  const int iterations = 2;
  const int sleeptime = 10; 

  struct thread_arg thread_pairs[THREAD_PAIRS];
  for(i=0; i<THREAD_PAIRS ; i++ ){
    thread_pairs[i].send_socket = -1;
    thread_pairs[i].send_port = 10000 + i;
    thread_pairs[i].recv_socket = -1;
    thread_pairs[i].recv_port = 20000 + i;  
    thread_pairs[i].sem = sys_sem_new(0);
    thread_pairs[i].packets = packets;
    thread_pairs[i].iterations = iterations;
    if(i==0){
      thread_pairs[i].send_tos = IPTOS_PREC_PRIORITY;
      thread_pairs[i].recv_tos = IPTOS_PREC_PRIORITY;    
    }
    else {
      thread_pairs[i].send_tos = IPTOS_PREC_ROUTINE;
      thread_pairs[i].recv_tos = IPTOS_PREC_ROUTINE;        
    }
  }

  init(); 

  for(i=0; i<THREAD_PAIRS ; i++ ){
    sys_thread_new("socket_thread", socket_thread, (void *)&thread_pairs[i], DEFAULT_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
    printf("%s: Main: Waiting for thread pair %d init.\n", name, i);
    sys_sem_wait(thread_pairs[i].sem);
  }

  for(i=(THREAD_PAIRS-1); i>=0 ; i-- ){  
    printf("%s: Main: Signalling thread pair %d to begin transmission.\n", name, i);
    sys_sem_signal(thread_pairs[i].sem);
  }

  printf("%s: Main: Integrated test will timeout after %d seconds.\n\n", name, sleeptime);
  sleep(sleeptime);
  printf("%s: Main: Test timeout.\n", name);

  stats_display();

  return 0;
}


